import re

regex = "(?:(FROM|JOIN|INTO|USING)\s+)(\w+)[.](\w+)"


def replace_query_details(query, schema_suffix_to_add):
    a = re.finditer(regex, query, flags=re.I)
    details = []
    template_detail = 'Found in position %s: %s.%s --> %s replaced with %s'
    for i in a:
        pos = i.span()
        group = i.groups()
        t = template_detail % (pos, group[1], group[2], group[1], group[1] + '_' + schema_suffix_to_add)
        details.append(t)
    return re.sub(regex, r'\1 \2' + '_' + schema_suffix_to_add + r'.\3', query, flags=re.I | re.VERBOSE), details



import datetime
import uuid


def generate_dck_info(row_data, batch_id=None, id_info='dck'):
    result = []
    _now = datetime.datetime.now()
    for row in row_data:
        if not row.get(f'__{id_info}_id___'):
            row[f'__{id_info}_id___'] = str(uuid.uuid4())
        if not row.get(f'__{id_info}_inserted_at___'):
            row[f'__{id_info}_inserted_at___'] = _now
        if batch_id and not row.get(f'__{id_info}_batch_id___'):
            row[f'__{id_info}_batch_id___'] = batch_id
        result.append(row)
    return result



import copy
import datetime
import json
import os
import random

import re
import time

import requests

from dbstream.tools.parse_data import treat_json_data
from dbstream.tools.regex import replace_query_details

from google.api_core.exceptions import Forbidden

class DBStream:

    def __init__(self, instance_name, client_id, special_env=None, id_info='dck'):
        self.instance_name = instance_name
        self.instance_type_prefix = ""
        self.client_id = client_id
        unix_time = time.mktime(datetime.datetime.now().timetuple())
        self.dbstream_instance_id = 'df-' + str(unix_time) + '-' + str(random.randint(1000, 9999))
        self.special_env = special_env if special_env else None
        self.id_info = id_info

    def error_if_function_not_exist(self, function_name):
        raise Exception("Function %s is not defined for %s" % (function_name, type(self).__name__))

    def prefix(self):
        return self.instance_type_prefix + "_" + self.instance_name

    def remote_host(self):
        return os.environ[self.prefix() + "_HOST"]

    def remote_port(self):
        return os.environ[self.prefix() + "_PORT"]

    def credentials(self):
        host = self.remote_host()
        port = self.remote_port()
        return {
            'database': os.environ[self.prefix() + "_DATABASE"],
            'user': os.environ[self.prefix() + "_USERNAME"],
            'host': host,
            'port': port,
            'password': os.environ[self.prefix() + "_PASSWORD"],
        }

    def _execute_query_custom(self, query) -> dict:
        self.error_if_function_not_exist("_execute_query_custom")

    def execute_query(self, query, apply_special_env=True):
        query = re.sub(' +', ' ', query)
        query = re.sub(' +\n', '\n', query)
        if apply_special_env and self.special_env:
            query = replace_query_details(query, self.special_env)[0]
        result = self._execute_query_custom(query)
        if isinstance(result, dict):
            if result.get('execute_query'):
                query_create_table = result.get('execute_query').group(0)
                if len(query_create_table.split('.')) > 1:
                    schema_name = query_create_table.split('.')[0]
                    table_name = query_create_table.split('.')[1]
                else:
                    schema_name = None
                    table_name = query_create_table.split('.')[0]

                url = os.environ.get("MONITORING_URL")
                if url:
                    body = {
                        "dbstream_instance_id": self.dbstream_instance_id,
                        "instance_name": self.instance_name,
                        "client_id": self.client_id,
                        "instance_type_prefix": self.instance_type_prefix,
                        "timestamp": str(datetime.datetime.now()),
                        "local_absolute_path": os.getcwd(),
                        "execute_query": True,
                        'schema_name': schema_name,
                        'table_name': table_name
                    }
                    r = requests.post(url=url, data=json.dumps(body))
                    print(r.status_code)
            return None
        return result

    def _send_data_custom(self, data, replace, **kwargs):
        pass

    def _send(self, **args):
        pass

    def send_data(self, data, replace=True, apply_special_env=True, **kwargs):
        data["columns_name"] = [c.replace(".", "_").replace("(", "_").replace(")", "_") for c in data["columns_name"]]
        data_copy = copy.deepcopy(data)
        if apply_special_env and self.special_env:
            schema_name = data_copy["table_name"].split(".")[0]
            table_name = data_copy["table_name"].split(".")[1]
            data_copy["table_name"] = "%s_%s.%s" % (schema_name, self.special_env, table_name)
            if kwargs.get("other_table_to_update"):
                schema_name = kwargs["other_table_to_update"].split(".")[0]
                table_name = kwargs["other_table_to_update"].split(".")[1]
                kwargs["other_table_to_update"] = "%s_%s.%s" % (schema_name, self.special_env, table_name)

        data_copy2 = copy.deepcopy(data_copy)
        if self._send_data_custom(data_copy, replace, **kwargs) != 0:
            url = os.environ.get("MONITORING_URL")
            if url:
                table_schema_name = data_copy2["table_name"].split(".")
                body = {
                    "dbstream_instance_id": self.dbstream_instance_id,
                    "instance_name": self.instance_name,
                    "client_id": self.client_id,
                    "instance_type_prefix": self.instance_type_prefix,
                    "schema_name": table_schema_name[0],
                    "table_name": table_schema_name[1],
                    "nb_rows": len(data_copy2["rows"]),
                    "nb_columns": len(data_copy2["columns_name"]),
                    "timestamp": str(datetime.datetime.now()),
                    "local_absolute_path": os.getcwd(),
                    "replace": replace
                }
                r = requests.post(url=url, data=json.dumps(body))
                print(r.status_code)

    def send(self, data, replace=False, apply_special_env=True, delay=5, batch_id=None, **kwargs):
        # data['data'] = generate_dck_info(data['data'])
        list_of_tables_to_send, list_of_pop_fields = treat_json_data(data, batch_id=batch_id, id_info=self.id_info)
        for d in list_of_tables_to_send:
            # if d.get('table_name') == 'test.test_orders_fulfillments_line_items':
            #     print(d.get('data'))
            # else:
            #     print('OK')
            column_names = []
            for r in d['data']:
                for k in r.keys():
                    if k in ('authorization', 'default'):
                        k = k + '_'
                    if k not in column_names:
                        column_names.append(k)
            data_to_send = {
                "columns_name": column_names,
                "rows": [[r.get(c) for c in column_names] for r in d['data']],
                "table_name": d.get('table_name')
            }
            if len(data_to_send['columns_name']) > 0:
                try:
                    self.send_data(data=data_to_send, replace=replace, apply_special_env=apply_special_env, **kwargs)
                except Forbidden:
                    time.sleep(delay)
                    self.send_data(data=data_to_send, replace=replace, apply_special_env=apply_special_env, **kwargs)

    def send_temp_data(self, data, schema_prefix, table, column_names, apply_special_env=True):
        data_to_send = {
            "columns_name": column_names,
            "rows": [[r.get(c) for c in column_names] for r in data],
            "table_name": schema_prefix + '.' + table + '_temp'}
        self.send_data(
            data=data_to_send,
            other_table_to_update=schema_prefix + '.' + table,
            replace=False,
            apply_special_env=apply_special_env)

    def clean(self, selecting_id, schema_prefix, table):
        self.error_if_function_not_exist("clean")

    def send_with_temp_table(self, data, column_names, selecting_id, schema_prefix, table):
        self.send_temp_data(data, schema_prefix, table, column_names)
        self.clean(selecting_id, schema_prefix, table)

    def get_max(self, schema, table, field, filter_clause):
        self.error_if_function_not_exist("get_max")

    def get_data_type(self, table_name, schema_name):
        self.error_if_function_not_exist("get_data_type")

    def create_view_from_columns(self, view_name, columns, schema_name, table_name):
        self.error_if_function_not_exist("create_view_from_columns")

    def create_schema(self, schema_name):
        self.error_if_function_not_exist("create_schema")

    def drop_schema(self, schema_name):
        self.error_if_function_not_exist("drop_schema")

    def build_pydatasource_view(self, query_string):
        self.error_if_function_not_exist("build_pydatasource_view")

    def build_pydatasource_table(self, query_string):
        self.error_if_function_not_exist("build_pydatasource_table")

    def build_pydatasource_table_cascade(self, query_string):
        self.error_if_function_not_exist("build_pydatasource_table_cascade")



import json

from dbstream.tools.dck_infos import generate_dck_info


def unest_data(row_data, k):
    result = []
    for r in row_data:
        if r.get(k):
            for rk in r[k].keys():
                r[k + '_' + rk] = r[k].get(rk)
        r.pop(k, None)
        result.append(r)
    return result


def treat_json_data(data, list_of_tables_to_send=None, list_of_pop_fields=None, batch_id=None, id_info='dck'):
    table_name = data['table_name']
    row_data = data['data']
    data['data'] = generate_dck_info(row_data, batch_id=batch_id, id_info=id_info)
    if not list_of_tables_to_send:
        list_of_tables_to_send = []
    if not list_of_pop_fields:
        list_of_pop_fields = {table_name: []}
    if not list_of_pop_fields.get(table_name):
        list_of_pop_fields[table_name] = []
    for row in row_data:
        row_keys = list(row.keys())
        for k in row_keys:
            k_table_name = table_name + '_' + k
            if k_table_name in [t['table_name'] for t in list_of_tables_to_send]:
                continue
            if k in list_of_pop_fields[table_name]:
                continue
            if isinstance(row[k], str):
                try:
                    if isinstance(json.loads(row[k]), dict) or isinstance(json.loads(row[k]), list):
                        for i in range(len(row_data)):
                            try:
                                row_data[i][k] = json.loads(row_data[i][k])
                            except TypeError:
                                pass
                except ValueError:
                    pass
            if isinstance(row[k], dict):
                row_data = unest_data(row_data, k)
                list_of_pop_fields[table_name].append(k)
                list_of_tables_to_send, list_of_pop_fields = treat_json_data(
                    {'table_name': table_name, 'data': row_data},
                    list_of_tables_to_send=list_of_tables_to_send,
                    list_of_pop_fields=list_of_pop_fields,
                    batch_id=batch_id,
                    id_info=id_info
                )
            elif isinstance(row[k], list):
                k_row_data = []
                for r in row_data:
                    if r.get(k) is not None:
                        for i in range(len(r[k])):
                            rr = r[k][i]
                            if not isinstance(rr, dict):
                                rr = {'value': rr}
                            rr['__' + table_name.split('.')[1] + f'__{id_info}_id___'] = r[f'__{id_info}_id___']
                            rr['__' + table_name.split('.')[1] + f'__{id_info}_id___' + 'order'] = i
                            k_row_data.append(rr)
                        r.pop(k, None)
                list_of_pop_fields[table_name].append(k)
                k_row_data = generate_dck_info(k_row_data, batch_id=batch_id, id_info=id_info)
                k_data = {'table_name': k_table_name, 'data': k_row_data}
                list_of_tables_to_send.append(k_data)
                list_of_tables_to_send, list_of_pop_fields = treat_json_data(k_data,
                                                                             list_of_tables_to_send=list_of_tables_to_send,
                                                                             list_of_pop_fields=list_of_pop_fields,
                                                                             batch_id=batch_id,
                                                                             id_info=id_info)
    if not list_of_tables_to_send:
        list_of_tables_to_send = [{'table_name': table_name, 'data': row_data}]
    elif table_name not in [t['table_name'] for t in list_of_tables_to_send]:
        list_of_tables_to_send.append({'table_name': table_name, 'data': row_data})
    return list_of_tables_to_send, list_of_pop_fields
